# frozen_string_literal: true

module Heptapod::NativeMercurial
  # Native Mercurial Migrator
  #
  # This is responsible for migrating a mercurial project from hg_git backend to native hg.
  class Migrator
    SHA_MAX_LEN = 40
    SHA_REGEX = '\b[0-9a-f]{7,40}\b'

    def initialize(project)
      @hg_sha_map = project.repository.hg_sha_map
      @project = project
    end

    def vcs_type_enforcer
      proj_initial_vcs = @project.vcs_type
      if @project.vcs_type == 'hg'
        @project.vcs_type = 'hg_git'
        @project.save!
      end

      yield
    ensure
      # ensure that native project remains native in case of errors
      if proj_initial_vcs == 'hg'
        @project.vcs_type = 'hg'
        @project.save!
      end
    end

    def migrate!
      vcs_type_enforcer do
        do_migrate
      end
    end

    def do_migrate
      Gitlab::AppLogger.info(
        message: "Starting native hg migration",
        project_id: @project.id,
        project_full_path: @project.full_path
      )

      migrate_issues
      migrate_merge_requests
      migrate_notes
      migrate_releases

      # ensure all gitlab specific state files are present
      output, status = @project.repository.hg_call(
        ["hpd-ensure-all-gitlab-specific-state-files"], force_system_user: true
      )
      raise StandardError, output.strip unless status == 0

      @project.vcs_type = 'hg'
      @project.save!

    rescue GRPC::Unavailable => err
      Gitlab::AppLogger.error(
        message: "Error: Could not connect to Gitaly, make sure it is running.",
        backtrace: Gitlab::BacktraceCleaner.clean_backtrace(err.backtrace),
        project_id: @project.id,
        project_full_path: @project.full_path
      )
      raise
    rescue StandardError => err
      Gitlab::AppLogger.error(
        message: "Native mercurial migration failed: #{err}",
        backtrace: Gitlab::BacktraceCleaner.clean_backtrace(err.backtrace),
        project_id: @project.id,
        project_full_path: @project.full_path
      )
      raise
    end

    def migrate_issues
      @project.issues.each do |issue|
        issue.title = replace_git_shas_with_hg(issue.title)
        issue.title_html = nil
        issue.description = replace_git_shas_with_hg(issue.description)
        issue.description_html = nil
        issue.save!
      end
    end

    def migrate_notes
      notes_service = Notes::BaseService.new(@project)
      @project.notes.each do |note|
        note.note = replace_git_shas_with_hg(note.note)
        migrate_note_positions(note)
        note.commit_id = replace_git_shas_with_hg(note.commit_id)
        note.note_html = nil
        # clear noteable diffs cache
        notes_service.clear_noteable_diffs_cache(note)
        note.save!
      end
    end

    def migrate_note_positions(note)
      if note.original_position.instance_of? Gitlab::Diff::Position
        note.original_position = _migrated_position(note.original_position)
      end

      if note.position.instance_of? Gitlab::Diff::Position
        note.position = _migrated_position(note.position)
      end

      if note.change_position.instance_of? Gitlab::Diff::Position
        note.change_position = _migrated_position(note.change_position)
      end
    end

    def _migrated_position(pos)
      pos_attrs = pos.to_h
      pos_attrs[:base_sha] = replace_git_shas_with_hg(pos_attrs[:base_sha])
      pos_attrs[:start_sha] = replace_git_shas_with_hg(
        pos_attrs[:start_sha])
      pos_attrs[:head_sha] = replace_git_shas_with_hg(pos_attrs[:head_sha])
      Gitlab::Diff::Position.new(pos_attrs)
    end

    def migrate_releases
      @project.releases.each do |rl|
        rl.sha = replace_git_shas_with_hg(rl.sha)
        rl.description = replace_git_shas_with_hg(rl.description)
        rl.description_html = nil
        rl.save!
      end
    end

    def migrate_merge_requests
      @project.merge_requests.each do |mr|
        mr.title = replace_git_shas_with_hg(mr.title)
        mr.title_html = nil
        mr.description = replace_git_shas_with_hg(mr.description)
        mr.description_html = nil

        # update all the columns ending with '_sha'
        mr.merge_commit_sha = replace_git_shas_with_hg(mr.merge_commit_sha)
        mr.rebase_commit_sha = replace_git_shas_with_hg(mr.rebase_commit_sha)
        mr.in_progress_merge_commit_sha = (
          replace_git_shas_with_hg(mr.in_progress_merge_commit_sha)
        )
        mr.squash_commit_sha = replace_git_shas_with_hg(mr.squash_commit_sha)
        mr.merge_ref_sha = replace_git_shas_with_hg(mr.merge_ref_sha)
        migrate_merge_request_diffs(mr)
        mr.save!
      end
    end

    def migrate_merge_request_diffs(mr)
      mr.merge_request_diffs.each do |diff|
        diff.base_commit_sha = replace_git_shas_with_hg(diff.base_commit_sha)
        diff.head_commit_sha = replace_git_shas_with_hg(diff.head_commit_sha)
        diff.start_commit_sha = replace_git_shas_with_hg(diff.start_commit_sha)
        diff.save!
      end
      # clear merge request diffs cache
      mr.diffs.clear_cache
    end

    private

    def replace_git_shas_with_hg(text)
      return unless text.present? && @hg_sha_map.present?

      git_shas = @hg_sha_map.keys
      text.gsub(/#{SHA_REGEX}/) do |match|
        if match.length == SHA_MAX_LEN && git_shas.include?(match)
          @hg_sha_map[match]
        else
          full_git_sha = lambda {
            git_shas.each do |sha|
              if sha.starts_with? match
                return sha
              end
            end
            nil
          }.call
          full_git_sha.nil? ? match : @hg_sha_map[full_git_sha]
        end
      end
    end
  end
end
